@@ -36,6 +36,41 @@
 #include "queue_track.h"
 #endif
 
+static int find_debugfs_mount_point(char *mount_path_buf, size_t buf_size);
+
+/*
+ * Helper function to resize buffer when needed.
+ * Returns 0 on success, -1 on failure.
+ */
+int resize_buffer_if_needed(char **buffer, size_t *current_size)
+{
+	if (config_buffer_size == *current_size)
+		return 0;
+
+	char *old_buffer = *buffer;
+	*buffer = realloc(*buffer, config_buffer_size);
+	if (!*buffer) {
+		warn("fail to increase the buffer... continue");
+		*buffer = old_buffer;
+		return -1;
+	}
+
+	*current_size = config_buffer_size;
+	return 0;
+}
+
+/*
+ * Helper function for memory allocation with error checking.
+ * Dies with an error message if allocation fails.
+ */
+void *allocate_memory(size_t count, size_t size)
+{
+	void *ptr = calloc(count, size);
+	if (!ptr)
+		die("Cannot allocate memory");
+	return ptr;
+}
+
 /*
  * fill_process_comm - process name from task group ID.
  */
@@ -216,14 +251,10 @@ int setup_signal_handling(void)
  */
 void __die(const char *fmt, ...)
 {
-	volatile int zero = 0;
-	int ret = errno;
 	va_list ap;
 
 	if (errno)
 		perror("stalld");
-	else
-		ret = -1;
 
 	va_start(ap, fmt);
 	fprintf(stderr, "  ");
@@ -232,11 +263,11 @@ void __die(const char *fmt, ...)
 
 	fprintf(stderr, "\n");
 
-	/* Die with a division by zero to keep the stack on GDB. */
+	/* in verbose mode, abort() to generate a crash dump */
 	if (config_verbose)
-		zero = 10 / zero;
+		abort();
 
-	exit(ret);
+	exit(EXIT_FAILURE);
 }
 
 /*
@@ -407,44 +438,6 @@ void cleanup_regex(unsigned int *nr_task, regex_t **compiled_expr)
 #define _STR(x) #x
 #define STR(x) _STR(x)
 
-static int find_mount(const char *mount, char *debugfs)
-{
-	char type[100];
-	FILE *fp;
-
-	if ((fp = fopen("/proc/mounts","r")) == NULL)
-		return 0;
-
-	while (fscanf(fp, "%*s %"
-		      STR(MAX_PATH)
-		      "s %99s %*s %*d %*d\n",
-		      debugfs, type) == 2) {
-		if (strcmp(type, mount) == 0)
-			break;
-	}
-	fclose(fp);
-
-	if (strcmp(type, mount) != 0)
-		return 0;
-	return 1;
-}
-
-static const char *find_debugfs(void)
-{
-	static char debugfs[MAX_DIR_PATH];
-	static int debugfs_found;
-
-	if (debugfs_found)
-		return debugfs;
-
-	if (!find_mount("debugfs", debugfs))
-		return "";
-
-	debugfs_found = 1;
-
-	return debugfs;
-}
-
 /*
  * Return true if the file at *path exists.
  */
@@ -465,13 +458,12 @@ static int check_file_exists(char *path)
  */
 static int find_debugfs_sched_debug(void)
 {
-	const char *debugfs;
+	char debugfs[MAX_DIR_PATH];
 	char *path;
 	int found;
 	size_t size;
 
-	debugfs = find_debugfs();
-	if (!debugfs)
+	if(find_debugfs_mount_point(debugfs, sizeof(debugfs)))
 		return 0;
 
 	size = strlen(debugfs) + strlen("sched/debug") + 2;
@@ -569,11 +561,10 @@ out_err:
 
 static int fill_sched_features_path(char *path, int path_size)
 {
-	const char *debugfs;
+	char debugfs[MAX_DIR_PATH];
 	int retval;
 
-	debugfs = find_debugfs();
-	if (strlen(debugfs) == 0)
+	if (find_debugfs_mount_point(debugfs, sizeof(debugfs)))
 		return 0;
 
 	snprintf(path, path_size, "%s/sched/features", debugfs);
@@ -776,6 +767,7 @@ static void print_usage(void)
 		"                               starving_threshold time, dispatch a specialized thread to monitor",
 		"                               it.",
 		"	   -O/--power_mode: works as a single threaded tool. Saves CPU, but loses precision.",
+		"	   -N/--no_idle_detect: disable idle CPU detection optimization (for testing)",
 		"	   -g/--granularity: set the granularity at which stalld checks for starving threads",
 		"	   -R/--reservation: percentage of CPU time reserved to stalld using SCHED_DEADLINE.",
 		"	   -a/--affinity: limit stalld's affinity",
@@ -1005,6 +997,7 @@ int parse_args(int argc, char **argv)
 			{"aggressive_mode",	no_argument,	   0, 'A'},
 			{"power_mode",		no_argument,	   0, 'O'},
 			{"adaptive_mode",	no_argument,	   0, 'M'},
+			{"no_idle_detect",	no_argument,	   0, 'N'},
 			{"help",		no_argument,	   0, 'h'},
 			{"boost_period",	required_argument, 0, 'p'},
 			{"boost_runtime",	required_argument, 0, 'r'},
@@ -1026,7 +1019,7 @@ int parse_args(int argc, char **argv)
 		/* getopt_long stores the option index here. */
 		int option_index = 0;
 
-		c = getopt_long(argc, argv, "lvkfAOMhsp:r:d:t:c:FVSg:i:I:R:b:a:",
+		c = getopt_long(argc, argv, "lvkfAOMNhsp:r:d:t:c:FVSg:i:I:R:b:a:",
 				 long_options, &option_index);
 
 		/* Detect the end of the options. */
@@ -1147,6 +1140,10 @@ int parse_args(int argc, char **argv)
 			config_single_threaded = 0;
 			config_aggressive = 0;
 			break;
+		case 'N':
+			config_idle_detection = 0;
+			log_msg("idle detection disabled\n");
+			break;
 		case 'R':
 			config_reservation = get_long_from_str(optarg);
 			if (config_reservation < 10 || config_reservation > 90)
@@ -1278,8 +1275,8 @@ static int find_debugfs_mount_point(char *mount_path_buf, size_t buf_size) {
  * Returns -1 if debugfs is not mounted, or other system error occurs.
  */
 int check_dl_server_dir_exists(void) {
-	char debugfs_mount_path[256];
-	char full_target_path[512];
+	char debugfs_mount_path[MAX_DIR_PATH];
+	char full_target_path[MAX_PATH];
 	struct stat st;
 	const char *dl_server_subdir = "/sched/fair_server";
 
@@ -1306,3 +1303,92 @@ int check_dl_server_dir_exists(void) {
 	log_msg("DL-server not detected: %s\n", strerror(errno));
 	return -1;
 }
+
+/*
+ * API to fetch the process group ID for a thread/process.
+ */
+int get_tgid(int pid)
+{
+	const char tgid_field[] = "Tgid:";
+	char file_location[PROC_PID_FILE_PATH_LEN];
+	char *status = NULL;
+	int tgid, n;
+	FILE *fp;
+
+	status = calloc(TMP_BUFFER_SIZE, sizeof(char));
+	if (status == NULL) {
+		return -ENOMEM;
+	}
+
+	n = sprintf(file_location, "/proc/%d/status", pid);
+	if (n < 0)
+		goto out_free_mem;
+
+	if ((fp = fopen(file_location, "r")) == NULL)
+		goto out_free_mem;
+
+	/* Iterate till we find the tgid field. */
+	while (1) {
+		if (fgets(status, TMP_BUFFER_SIZE, fp) == NULL)
+			goto out_close_fd;
+		if (!(strncmp(status, tgid_field, sizeof(tgid_field) - 1)))
+			break;
+		/*
+		 * Zero out the buffer just in case
+		 */
+		memset(status, 0, TMP_BUFFER_SIZE);
+	}
+	/*
+	 * Since we're now at the line we're interested in, let's read
+	 * in the field that we want.
+	 */
+	if (sscanf(status, "%*s %d", &tgid) != 1)
+		goto out_close_fd;
+
+	fclose(fp);
+	free(status);
+	return tgid;
+
+out_close_fd:
+	fclose(fp);
+out_free_mem:
+	free(status);
+	return -EINVAL;
+}
+
+/*
+ * Read the content of /proc/stat into the input buffer.
+ * Used by functions doing cpu idle detection
+ */
+int read_proc_stat(char *buffer, int size)
+{
+	int position = 0;
+	int retval;
+	int fd;
+
+	fd = open("/proc/stat", O_RDONLY);
+
+	if (fd < 0)
+		goto out_error;
+
+	do {
+		retval = read(fd, &buffer[position], size - position);
+		if (retval < 0)
+			goto out_close_fd;
+
+		position += retval;
+
+	} while (retval > 0 && position < size);
+
+	buffer[position-1] = '\0';
+
+	close(fd);
+
+	return position;
+
+out_close_fd:
+	close(fd);
+
+out_error:
+	return 0;
+}
